/*
 *
 * Function for recognizing partitionability
 * (for pure/nonpure d-dimensional complexes with d<= 3)
 *
 * by M.Hachimori, Jul 2006.
 */

#include <stdio.h>
#include <stdlib.h>
#include "s_complex.h"
/* #define MAXVERT 100 */

int m_empty;
int m_vert[MAXVERT];
int m_edge[MAXVERT][MAXVERT];
int m_tri[MAXVERT][MAXVERT][MAXVERT];
int c_empty;
int c_vert[MAXVERT];
int c_edge[MAXVERT][MAXVERT];
int c_tri[MAXVERT][MAXVERT][MAXVERT];
int f_vec[MAXDIM], h_vec[MAXDIM];


int l[4], ind[4];

void cancel_np3(Simplices* facets, int num);
int check_np3(Simplices* facets, int num);
void putpart_np3(Simplices* facets, int num);
void printdata_np3(Simplices* facets);

/*
 * Function for recognizing partitionability
 * (for pure/nonpure d-dimensional complexes with d<= 3)
 *   Input: list of facets
 *   return 1 if partitionable
 *   return 0 if not partitionable
 */
int is_partitionable_nonpure3(facets)
  Simplices *facets;
{

  Simplices *faces, *tmp_faces;
  Simplices *tmp_facets;
  int i, j, k, dim, d;

  /* Check the dimension of the complex.*/
  dim = -1;
  tmp_facets = facets;
  while(tmp_facets != NULL) {
    if (tmp_facets->vert[0] > dim) dim = tmp_facets->vert[0];
    if (dim > 3) {
      printf("There are facets of dimension > 3. Error.\n");
      return(-1);
    }
    tmp_facets = tmp_facets->next;
  }



  /* initialization */
  /* c_xxx indicate whether faces are covered or not */
  for (i = 0; i < MAXVERT; i ++) {
    c_vert[i] = 0;
    for (j = 0; j < MAXVERT; j ++) {
      c_edge[i][j] = 0;
      for (k = 0; k < MAXVERT; k ++)
        c_tri[i][j][k] = 0;
      }
    }
  c_empty = 0;
  /* m_xxx indicate the multiplicity of incident facets */
  for (i = 0; i < MAXVERT; i ++) {
    m_vert[i] = 0;
    for (j = 0; j < MAXVERT; j ++) {
      m_edge[i][j] = 0; 
      for (k = 0; k < MAXVERT; k ++)
        m_tri[i][j][k] = 0;
      }
    }
  m_empty = 0;
  tmp_facets = facets;
  while(tmp_facets != NULL) {
    if (tmp_facets->vert[0] == 3) {
      m_vert[tmp_facets->vert[1]] ++;
      m_vert[tmp_facets->vert[2]] ++;
      m_vert[tmp_facets->vert[3]] ++;
      m_vert[tmp_facets->vert[4]] ++;
      m_edge[tmp_facets->vert[1]][tmp_facets->vert[2]] ++;
      m_edge[tmp_facets->vert[1]][tmp_facets->vert[3]] ++;
      m_edge[tmp_facets->vert[1]][tmp_facets->vert[4]] ++;
      m_edge[tmp_facets->vert[2]][tmp_facets->vert[3]] ++;
      m_edge[tmp_facets->vert[2]][tmp_facets->vert[4]] ++;
      m_edge[tmp_facets->vert[3]][tmp_facets->vert[4]] ++;
      m_tri[tmp_facets->vert[1]][tmp_facets->vert[2]][tmp_facets->vert[3]] ++;
      m_tri[tmp_facets->vert[1]][tmp_facets->vert[2]][tmp_facets->vert[4]] ++;
      m_tri[tmp_facets->vert[1]][tmp_facets->vert[3]][tmp_facets->vert[4]] ++;
      m_tri[tmp_facets->vert[2]][tmp_facets->vert[3]][tmp_facets->vert[4]] ++;
      m_empty ++;
      tmp_facets->vert[5] = -1;  /* facets->vert[5] indicates the type of interval
                                 in a partition. 
                                   0: [empty,facet]
                                   1,2,3,4 : [1,f],[2,f],[3,f],[4,f]
                                   5,6,7,8,9,10 : [12,f],[13,f],[14,f],[23,f],[24,f],[34,f]
                                   11,12,13,14: [123,f],[124,f],[134,f],[234,f]
                                   15: [f,f]
                                   -1: not assigned yet.  */
      }
    else if (tmp_facets->vert[0] == 2) {
      m_vert[tmp_facets->vert[1]] ++;
      m_vert[tmp_facets->vert[2]] ++;
      m_vert[tmp_facets->vert[3]] ++;
      m_edge[tmp_facets->vert[1]][tmp_facets->vert[2]] ++;
      m_edge[tmp_facets->vert[1]][tmp_facets->vert[3]] ++;
      m_edge[tmp_facets->vert[2]][tmp_facets->vert[3]] ++;
      m_empty ++;
      tmp_facets->vert[4] = -1;  /* facets->vert[4] indicates the type of interval
                                   in a partition. 
                                     0: [empty,facet]
                                     1,2,3 : [1,facet], [2,facet], [3,facet]
                                     4,5,6 : [12,facet], [13,facet], [23,facet]
                                     7: [facet,facet]
                                     -1: not assigned yet.  */
      }
    else if (tmp_facets->vert[0] == 1) {
    m_vert[tmp_facets->vert[1]] ++;
    m_vert[tmp_facets->vert[2]] ++;
    m_empty ++;
    tmp_facets->vert[3] = -1;  /* facets->vert[3] indicates the type of interval
                                 in a partition. 
                                   0: [empty,facet]
                                   1,2: [1,facet], [2,facet]
                                   3: [facet,facet]
                                   -1: not assigned yet.  */
      }
    else /* if (tmp_facets->vert[0] == 0)*/ {
    m_vert[tmp_facets->vert[1]] ++;
    m_empty ++;
    tmp_facets->vert[2] = -1;  /* facets->vert[2] indicates the type of interval
                                 in a partition. 
                                   0: [empty,facet]
                                   1: [facet,facet]
                                   -1: not assigned yet.  */
      }
    tmp_facets = tmp_facets->next;
  }
  l[0] = 1; l[1] = 3; l[2] = 7; l[3] = 15; /* largest num of type for each dim. */
  ind[0] = 2; ind[1] = 3; ind[2] = 4; ind[3] = 5; /* index of the type-flag for each dim. */

  /* main loop */
  tmp_facets = facets;
  while(tmp_facets != NULL) {
#ifdef DEBUG
    printdata_np3(facets);
#endif
    /* check types in the ordering -1,l,0,1,2,...,l-1 */
    d = tmp_facets->vert[0];
    if (tmp_facets->vert[ind[d]] == l[d]-1) {  /* backtrack */
      tmp_facets->vert[ind[d]] = -1;
      if (tmp_facets->prev == NULL) return 0;
      else {
        tmp_facets = tmp_facets->prev;
        d = tmp_facets->vert[0]; /* take d to the new dim of tmp_facets */
        cancel_np3(tmp_facets, tmp_facets->vert[ind[d]]);
        continue;
        }
      }
    else {
      if (tmp_facets->vert[ind[d]] == -1) tmp_facets->vert[ind[d]] = l[d];
      else if (tmp_facets->vert[ind[d]] == l[d]) tmp_facets->vert[ind[d]] = 0;
      else tmp_facets->vert[ind[d]] ++;
      if (check_np3(tmp_facets, tmp_facets->vert[ind[d]])) {
        putpart_np3(tmp_facets, tmp_facets->vert[ind[d]]);
        if (tmp_facets->next != NULL) {
          tmp_facets = tmp_facets->next;
          continue;
          }
        else  goto PARTITIONABLE;
        }

      }
    }

  return -1;

  PARTITIONABLE:
  printdata_np3(facets);
  return 1;

}

void cancel_np3(facets, num)
  Simplices* facets;
  int num;
{
  if (facets->vert[0] == 3) {
    m_empty ++;
    m_vert[facets->vert[1]] ++;
    m_vert[facets->vert[2]] ++;
    m_vert[facets->vert[3]] ++;
    m_vert[facets->vert[4]] ++;
    m_edge[facets->vert[1]][facets->vert[2]] ++;
    m_edge[facets->vert[1]][facets->vert[3]] ++;
    m_edge[facets->vert[1]][facets->vert[4]] ++;
    m_edge[facets->vert[2]][facets->vert[3]] ++;
    m_edge[facets->vert[2]][facets->vert[4]] ++;
    m_edge[facets->vert[3]][facets->vert[4]] ++;
    m_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
    m_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
    m_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
    m_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
    switch(num) {
      case 0:
        c_empty --;
        c_vert[facets->vert[1]] --;
        c_vert[facets->vert[2]] --;
        c_vert[facets->vert[3]] --;
        c_vert[facets->vert[4]] --;
        c_edge[facets->vert[1]][facets->vert[2]] --;
        c_edge[facets->vert[1]][facets->vert[3]] --;
        c_edge[facets->vert[1]][facets->vert[4]] --;
        c_edge[facets->vert[2]][facets->vert[3]] --;
        c_edge[facets->vert[2]][facets->vert[4]] --;
        c_edge[facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 1:
        c_vert[facets->vert[1]] --;
        c_edge[facets->vert[1]][facets->vert[2]] --;
        c_edge[facets->vert[1]][facets->vert[3]] --;
        c_edge[facets->vert[1]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 2:
        c_vert[facets->vert[2]] --;
        c_edge[facets->vert[1]][facets->vert[2]] --;
        c_edge[facets->vert[2]][facets->vert[3]] --;
        c_edge[facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 3:
        c_vert[facets->vert[3]] --;
        c_edge[facets->vert[1]][facets->vert[3]] --;
        c_edge[facets->vert[2]][facets->vert[3]] --;
        c_edge[facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 4:
        c_vert[facets->vert[4]] --;
        c_edge[facets->vert[1]][facets->vert[4]] --;
        c_edge[facets->vert[2]][facets->vert[4]] --;
        c_edge[facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 5:
        c_edge[facets->vert[1]][facets->vert[2]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        break;
      case 6:
        c_edge[facets->vert[1]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 7:
        c_edge[facets->vert[1]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 8:
        c_edge[facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 9:
        c_edge[facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 10:
        c_edge[facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 11:
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
        break;
      case 12:
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
        break;
      case 13:
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 14:
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
        break;
      case 15:
        break;
      }
    }
  else if (facets->vert[0] == 2) {
    m_empty ++;
    m_vert[facets->vert[1]] ++;
    m_vert[facets->vert[2]] ++;
    m_vert[facets->vert[3]] ++;
    m_edge[facets->vert[1]][facets->vert[2]] ++;
    m_edge[facets->vert[1]][facets->vert[3]] ++;
    m_edge[facets->vert[2]][facets->vert[3]] ++;
    switch(num) {
      case 0:
        c_empty --;
        c_vert[facets->vert[1]] --;
        c_vert[facets->vert[2]] --;
        c_vert[facets->vert[3]] --;
        c_edge[facets->vert[1]][facets->vert[2]] --;
        c_edge[facets->vert[1]][facets->vert[3]] --;
        c_edge[facets->vert[2]][facets->vert[3]] --;
        break;
      case 1:
        c_vert[facets->vert[1]] --;
        c_edge[facets->vert[1]][facets->vert[2]] --;
        c_edge[facets->vert[1]][facets->vert[3]] --;
        break;
      case 2:
        c_vert[facets->vert[2]] --;
        c_edge[facets->vert[1]][facets->vert[2]] --;
        c_edge[facets->vert[2]][facets->vert[3]] --;
        break;
      case 3:
        c_vert[facets->vert[3]] --;
        c_edge[facets->vert[1]][facets->vert[3]] --;
        c_edge[facets->vert[2]][facets->vert[3]] --;
        break;
      case 4:
        c_edge[facets->vert[1]][facets->vert[2]] --;
        break;
      case 5:
        c_edge[facets->vert[1]][facets->vert[3]] --;
        break;
      case 6:
        c_edge[facets->vert[2]][facets->vert[3]] --;
        break;
      case 7:
        break;
      }
    }
  else if (facets->vert[0] == 1) {
    m_empty ++;
    m_vert[facets->vert[1]] ++;
    m_vert[facets->vert[2]] ++;
    switch(num) {
      case 0:
        c_empty --;
        c_vert[facets->vert[1]] --;
        c_vert[facets->vert[2]] --;
        break;
      case 1:
        c_vert[facets->vert[1]] --;
        break;
      case 2:
        c_vert[facets->vert[2]] --;
        break;
      case 3:
        break;
      }
    }
  else {
    m_empty ++;
    switch(num) {
      case 0:
        c_empty --;
        break;
      case 1:
        break;
      }
    }
  
  return;

}

int check_np3(facets, num)
  Simplices* facets;
  int num;
{

  if (facets->vert[0] == 3) {
    /* check whether faces are already covered or not */
    switch(num) {
      case 0:
        if (c_empty == 1 ||
            c_vert[facets->vert[1]] == 1 ||
            c_vert[facets->vert[2]] == 1 ||
            c_vert[facets->vert[3]] == 1 ||
            c_vert[facets->vert[4]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
            c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 1:
        if (
            c_vert[facets->vert[1]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 2:
        if (
            c_vert[facets->vert[2]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 3:
        if (
            c_vert[facets->vert[3]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 4:
        if (
            c_vert[facets->vert[4]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
            c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
              c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 5:
          if (
            c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 6:
        if (
            c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 7:
        if (
            c_edge[facets->vert[1]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 8:
        if (
            c_edge[facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 ||
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 9:
        if (
            c_edge[facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 10:
        if (
            c_edge[facets->vert[3]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 ||
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 11:
        if (
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 12:
        if (
            c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 13:
        if (
            c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 14:
        if (
            c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 15:
        break;
    }


    /* if there is a face belonging only to the facet ... */
    if (c_empty == 0 && m_empty == 1 && num != 0) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[1]] == 0 && 
        m_vert[facets->vert[1]] == 1 &&
          !(num == 0 || num == 1) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[2]] == 0 && 
        m_vert[facets->vert[2]] == 1 &&
          !(num == 0 || num == 2) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[3]] == 0 && 
        m_vert[facets->vert[3]] == 1 &&
          !(num == 0 || num == 3) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[4]] == 0 && 
        m_vert[facets->vert[4]] == 1 &&
          !(num == 0 || num == 4) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[1]][facets->vert[2]] == 0 &&
        m_edge[facets->vert[1]][facets->vert[2]] == 1 &&
          !(num == 0 || num == 1 || num == 2 || num == 5) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[1]][facets->vert[3]] == 0 &&
        m_edge[facets->vert[1]][facets->vert[3]] == 1 &&
          !(num == 0 || num == 1 || num == 3 || num == 6) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[1]][facets->vert[4]] == 0 &&
        m_edge[facets->vert[1]][facets->vert[4]] == 1 &&
          !(num == 0 || num == 1 || num == 4 || num == 7) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[2]][facets->vert[3]] == 0 &&
        m_edge[facets->vert[2]][facets->vert[3]] == 1 &&
          !(num == 0 || num == 2 || num == 3 || num == 8) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[2]][facets->vert[4]] == 0 &&
        m_edge[facets->vert[2]][facets->vert[4]] == 1 &&
          !(num == 0 || num == 2 || num == 4 || num == 9) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[3]][facets->vert[4]] == 0 &&
        m_edge[facets->vert[3]][facets->vert[4]] == 1 &&
          !(num == 0 || num == 3 || num == 4 || num == 10) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 0 &&
        m_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] == 1 &&
          !(num == 0 || num == 1 || num == 2 || num == 3 || 
                        num == 5 || num == 6 || num == 8 || num == 11) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 0 &&
        m_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] == 1 &&
          !(num == 0 || num == 1 || num == 2 || num == 4 || 
                        num == 5 || num == 7 || num == 9 || num == 12) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 0 &&
        m_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] == 1 &&
          !(num == 0 || num == 1 || num == 3 || num == 4 || 
                        num == 6 || num == 7 || num == 10 || num == 13) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 0 &&
        m_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] == 1 &&
          !(num == 0 || num == 2 || num == 3 || num == 4 || 
                        num == 8 || num == 9 || num == 10 || num == 14) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    return 1;
    }


  else if (facets->vert[0] == 2) {
    /* check whether faces are already covered or not */
    switch(num) {
      case 0:
        if (c_empty == 1 ||
            c_vert[facets->vert[1]] == 1 ||
            c_vert[facets->vert[2]] == 1 ||
            c_vert[facets->vert[3]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[3]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 1:
        if (
            c_vert[facets->vert[1]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[3]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 2:
        if (
            c_vert[facets->vert[2]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[2]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[3]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 3:
        if (
            c_vert[facets->vert[3]] == 1 ||
            c_edge[facets->vert[1]][facets->vert[3]] == 1 ||
            c_edge[facets->vert[2]][facets->vert[3]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 4:
        if (
            c_edge[facets->vert[1]][facets->vert[2]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 5:
        if (
            c_edge[facets->vert[1]][facets->vert[3]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 6:
        if (
            c_edge[facets->vert[2]][facets->vert[3]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 7:
        break;
    }


    /* if there is a face belonging only to the facet ... */
    if (c_empty == 0 && m_empty == 1 && num != 0) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[1]] == 0 && 
        m_vert[facets->vert[1]] == 1 &&
          !(num == 0 || num == 1) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[2]] == 0 && 
        m_vert[facets->vert[2]] == 1 &&
          !(num == 0 || num == 2) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[3]] == 0 && 
        m_vert[facets->vert[3]] == 1 &&
          !(num == 0 || num == 3) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[1]][facets->vert[2]] == 0 &&
        m_edge[facets->vert[1]][facets->vert[2]] == 1 &&
          !(num == 0 || num == 1 || num == 2 || num == 4) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[1]][facets->vert[3]] == 0 &&
        m_edge[facets->vert[1]][facets->vert[3]] == 1 &&
          !(num == 0 || num == 1 || num == 3 || num == 5) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_edge[facets->vert[2]][facets->vert[3]] == 0 &&
        m_edge[facets->vert[2]][facets->vert[3]] == 1 &&
          !(num == 0 || num == 2 || num == 3 || num == 6) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    return 1;
    }

  else if (facets->vert[0] == 1) {
    /* check whether faces are already covered or not */
    switch(num) {
      case 0:
        if (c_empty == 1 ||
            c_vert[facets->vert[1]] == 1 ||
            c_vert[facets->vert[2]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 1:
        if (
            c_vert[facets->vert[1]] == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 2:
        if (
            c_vert[facets->vert[2]] == 1 
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 3:
        break;
      }

    /* if there is a face belonging only to the facet ... */
    if (c_empty == 0 && m_empty == 1 && num != 0) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[1]] == 0 && 
        m_vert[facets->vert[1]] == 1 &&
          !(num == 0 || num == 1) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }
    if (c_vert[facets->vert[2]] == 0 && 
        m_vert[facets->vert[2]] == 1 &&
          !(num == 0 || num == 2) ) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }

    }

  else {
    /* check whether faces are already covered or not */
    switch(num) {
      case 0:
        if (c_empty == 1
            ) {
#ifdef DEBUG
printf("multiple cover\n");
#endif
          return 0;
          }
        break;
      case 1:
        break;
      }

    /* if there is a face belonging only to the facet ... */
    if (c_empty == 0 && m_empty == 1 && num != 0) {
#ifdef DEBUG
printf("last facet forcing\n");
#endif
      return 0;
      }

    }


  return 1;

}

void putpart_np3(facets, num)
  Simplices* facets;
  int num;
{
  if (facets->vert[0] == 3) {
    m_empty --;
    m_vert[facets->vert[1]] --;
    m_vert[facets->vert[2]] --;
    m_vert[facets->vert[3]] --;
    m_vert[facets->vert[4]] --;
    m_edge[facets->vert[1]][facets->vert[2]] --;
    m_edge[facets->vert[1]][facets->vert[3]] --;
    m_edge[facets->vert[1]][facets->vert[4]] --;
    m_edge[facets->vert[2]][facets->vert[3]] --;
    m_edge[facets->vert[2]][facets->vert[4]] --;
    m_edge[facets->vert[3]][facets->vert[4]] --;
    m_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] --;
    m_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] --;
    m_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] --;
    m_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] --;
    switch(num) {
      case 0:
        c_empty ++;
        c_vert[facets->vert[1]] ++;
        c_vert[facets->vert[2]] ++;
        c_vert[facets->vert[3]] ++;
        c_vert[facets->vert[4]] ++;
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        c_edge[facets->vert[1]][facets->vert[4]] ++;
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        c_edge[facets->vert[2]][facets->vert[4]] ++;
        c_edge[facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 1:
        c_vert[facets->vert[1]] ++;
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        c_edge[facets->vert[1]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 2:
        c_vert[facets->vert[2]] ++;
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        c_edge[facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 3:
        c_vert[facets->vert[3]] ++;
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        c_edge[facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 4:
        c_vert[facets->vert[4]] ++;
        c_edge[facets->vert[1]][facets->vert[4]] ++;
        c_edge[facets->vert[2]][facets->vert[4]] ++;
        c_edge[facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 5:
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        break;
      case 6:
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 7:
        c_edge[facets->vert[1]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 8:
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 9:
        c_edge[facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 10:
        c_edge[facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 11:
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[3]] ++;
        break;
      case 12:
        c_tri[facets->vert[1]][facets->vert[2]][facets->vert[4]] ++;
        break;
      case 13:
        c_tri[facets->vert[1]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 14:
        c_tri[facets->vert[2]][facets->vert[3]][facets->vert[4]] ++;
        break;
      case 15:
        break;
      }
    }
  else if (facets->vert[0] == 2) {
    m_empty --;
    m_vert[facets->vert[1]] --;
    m_vert[facets->vert[2]] --;
    m_vert[facets->vert[3]] --;
    m_edge[facets->vert[1]][facets->vert[2]] --;
    m_edge[facets->vert[1]][facets->vert[3]] --;
    m_edge[facets->vert[2]][facets->vert[3]] --;
    switch(num) {
      case 0:
        c_empty ++;
        c_vert[facets->vert[1]] ++;
        c_vert[facets->vert[2]] ++;
        c_vert[facets->vert[3]] ++;
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        break;
      case 1:
        c_vert[facets->vert[1]] ++;
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        break;
      case 2:
        c_vert[facets->vert[2]] ++;
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        break;
      case 3:
        c_vert[facets->vert[3]] ++;
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        break;
      case 4:
        c_edge[facets->vert[1]][facets->vert[2]] ++;
        break;
      case 5:
        c_edge[facets->vert[1]][facets->vert[3]] ++;
        break;
      case 6:
        c_edge[facets->vert[2]][facets->vert[3]] ++;
        break;
      case 7:
        break;
      }
    }
  else if (facets->vert[0] == 1) {
    m_empty --;
    m_vert[facets->vert[1]] --;
    m_vert[facets->vert[2]] --;
    switch(num) {
      case 0:
        c_empty ++;
        c_vert[facets->vert[1]] ++;
        c_vert[facets->vert[2]] ++;
        break;
      case 1:
        c_vert[facets->vert[1]] ++;
        break;
      case 2:
        c_vert[facets->vert[2]] ++;
        break;
      case 3:
        break;
      }
    }
  else {
    m_empty --;
    switch(num) {
      case 0:
        c_empty ++;
        break;
      case 1:
        break;
      }
    }
  
  return;

}

void printdata_np3(facets)
  Simplices* facets;
{
  Simplices* tmp_facets;
  int i;

  tmp_facets = facets;
  while(tmp_facets != NULL) {
    for (i = 1; i <= tmp_facets->vert[0] + 1; i ++)
      printf("%d ", tmp_facets->vert[i]);
    printf(": ");
    if (tmp_facets->vert[0] == 3) {
      switch(tmp_facets->vert[5]) {
        case 0:
          printf("empty");
          break;
        case 1:
          printf("%d", tmp_facets->vert[1]);
          break;
        case 2:
          printf("%d", tmp_facets->vert[2]);
          break;
        case 3:
          printf("%d", tmp_facets->vert[3]);
          break;
        case 4:
          printf("%d", tmp_facets->vert[4]);
          break;
        case 5:
          printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[2]);
          break;
        case 6:
          printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[3]);
          break;
        case 7:
          printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[4]);
          break;
        case 8:
          printf("%d %d", tmp_facets->vert[2], tmp_facets->vert[3]);
          break;
        case 9:
          printf("%d %d", tmp_facets->vert[2], tmp_facets->vert[4]);
          break;
        case 10:
          printf("%d %d", tmp_facets->vert[3], tmp_facets->vert[4]);
          break;
        case 11:
          printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[3]);
          break;
        case 12:
          printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[4]);
          break;
        case 13:
          printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[3], tmp_facets->vert[4]);
          break;
        case 14:
          printf("%d %d %d", tmp_facets->vert[2], tmp_facets->vert[3], tmp_facets->vert[4]);
          break;
        case 15:
          printf("%d %d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[3], tmp_facets->vert[4]);
          break;
        }
      printf("\n");
      }
    else if (tmp_facets->vert[0] == 2) {
      switch(tmp_facets->vert[4]) {
        case 0:
          printf("empty");
          break;
        case 1:
          printf("%d", tmp_facets->vert[1]);
          break;
        case 2:
          printf("%d", tmp_facets->vert[2]);
          break;
        case 3:
          printf("%d", tmp_facets->vert[3]);
          break;
        case 4:
          printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[2]);
          break;
        case 5:
          printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[3]);
          break;
        case 6:
          printf("%d %d", tmp_facets->vert[2], tmp_facets->vert[3]);
          break;
        case 7:
          printf("%d %d %d", tmp_facets->vert[1], tmp_facets->vert[2], tmp_facets->vert[3]);
          break;
        }
      printf("\n");
      }
    else if (tmp_facets->vert[0] == 1) {
      switch(tmp_facets->vert[3]) {
        case 0:
          printf("empty");
          break;
        case 1:
          printf("%d", tmp_facets->vert[1]);
          break;
        case 2:
          printf("%d", tmp_facets->vert[2]);
          break;
        case 3:
          printf("%d %d", tmp_facets->vert[1], tmp_facets->vert[2]);
          break;
        }
      printf("\n");
      }
    else {
      switch(tmp_facets->vert[2]) {
        case 0:
          printf("empty");
          break;
        case 1:
          printf("%d", tmp_facets->vert[1]);
          break;
        }
      printf("\n");
      }
    tmp_facets = tmp_facets->next;
    }
  printf("\n");
  return;
}

