/*
 * header file for a program
 * to calculate simplicial complexes
 *
 * by M.Hachimori, May 2000
 */

#include <stdio.h>
#include <stdlib.h>

/*#########################################################################*/
/*
 * Some maximum numbers defined here.
 * Should be changed if larger data will be treated
 * in the computation.
 */
#define MAXDIM 20      /* max. of dimension of simplicial complexes */
#define MAXLINE 1000 /* max. of characters in one line in the input file  */
#define MAXVERT 200  /* max. of numbering to be used for vertices. */

/*#########################################################################*/

/* simplicial complex */
/* A face is represented by [dim, v_1, v_2, ..., v_(dim+1), X, X, ...],
   a vector with dim+2 entries.
 */  

struct simplices { short vert[MAXDIM + 2]; 
                   struct simplices *prev;
                   struct simplices *next; };
typedef struct simplices Simplices;      /* a set of simplices */

struct simplices_i { short vert[MAXDIM + 2];    /* a set of intervals. vert[] is the facet, and */
                     short subface[MAXDIM + 2]; /* subface[] indicates the lower face of the interval */
                   struct simplices_i *prev;
                   struct simplices_i *next; };
typedef struct simplices_i Simplices_i;      /* a set of simplices with a subface */


struct faces { int dim;
                 int n;
                 Simplices *faces; };
typedef struct faces Faces;              /* a set of simplices
                                            with additional information */

typedef struct list_simplices List;     /* a list of "set of simplices" */
struct list_simplices {Simplices *simplices;
                       List *next;};



/**********************************************
       prototypes of functions
 **********************************************/
/* in "io.c" */
int char_to_array(char *ch, short *ans);
void array_to_char(short *arr, char s[]);
Simplices *input_faces(char filename[]);
int cmp(short *a, short *b);
void print_faces(Simplices *simplices);

/* in "tools.c" */
void sort_faces(Simplices *simplices);
void msort(Simplices *simplices,
               int n);
int cmp_faces(short simplices1[], 
              short simplices2[]);
void delete_faces(Simplices *simplices);
int search(short simplex[], Simplices *simplices);
int binarysearch(short simplex[], Simplices *simplices, int n);
int face_contained(short simplex1[], short simplex2[]);

/* in "elementary.c" */
Simplices *facets2faces(Simplices *facets);
Simplices *faces2facets(Simplices *faces);
void f_to_h(int fvec[], int hvec[], int dim);
void get_fvector(Simplices *faces, int f_vec[]);


/* in "partitionable2.c", "partitionable3.c", 
      "partitionable_nonpure.c", "partitionable_nonpure-d.c */
int is_partitionable2(Simplices *facets);
int is_partitionable3(Simplices *facets);
int is_partitionable_nonpure3(Simplices *facets);
int is_partitionable_nonpure_d(Simplices *facets);

